# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: GPL-2.0-or-later
"""Unit tests for the `debsigs-autosign` tool."""

from __future__ import annotations

import shutil
import subprocess  # noqa: S404
import typing

import pytest

from . import util
from .util import test_cfg  # noqa: F401  # pytest fixture


if typing.TYPE_CHECKING:
    from typing import Final


def test_autosign(test_cfg: util.ConfigAggr) -> None:  # noqa: F811  # pytest fixture
    """Sign files in different ways."""
    if not test_cfg.has_all_features(
        {"cmd-list": "1", "cmd-sign": "1", "i-gpg": "1", "tool-autosign": "0.1"},
    ):
        pytest.skip("No `debsigs-autosign` or `debsigs --list` support")

    with test_cfg.cfg_with_tempd() as cfg:
        pub_key: Final = test_cfg.pub_key
        pkg_orig: Final = test_cfg.changes_orig.debs[0]

        pkg_signed: Final = cfg.path.work / pkg_orig.name
        shutil.copy2(pkg_orig, pkg_signed)
        assert util.debsigs_list(cfg, pkg_signed) == []

        env_with_path: Final = dict(cfg.env)
        env_with_path["PATH"] = f"{cfg.prog.debsigs.parent}:{env_with_path['PATH']}"

        subprocess.run(  # noqa: S603
            [cfg.prog.debsigs.parent / "debsigs-autosign", "maint"],
            check=True,
            cwd=cfg.path.work,
            encoding="UTF-8",
            env=env_with_path,
            input=f"{pkg_signed}\n",
        )
        lines_maint_only: Final = util.debsigs_list(cfg, pkg_signed)

        sign_key: Final = util.get_signing_subkey(pub_key)
        match lines_maint_only:
            case [single] if single.sig_type == "maint" and single.key_id == sign_key.key_id:
                pass

            case _:
                raise AssertionError(repr(lines_maint_only))

        # OK, let's do that again...
        subprocess.run(  # noqa: S603
            [cfg.prog.debsigs.parent / "debsigs-autosign", "origin"],
            check=True,
            cwd=cfg.path.work,
            encoding="UTF-8",
            env=env_with_path,
            input=f"{pkg_signed}\n",
        )
        lines_both: Final = util.debsigs_list(cfg, pkg_signed)

        match lines_both:
            case [
                maint,
                origin,
            ] if (
                maint.sig_type == "maint"
                and maint.key_id == sign_key.key_id
                and origin.sig_type == "origin"
                and origin.key_id == sign_key.key_id
            ):
                pass

            case _:
                raise AssertionError(repr(lines_both))



# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: GPL-2.0-or-later
"""Unit tests for the main functionality of the `debsigs` tool."""

from __future__ import annotations

import shutil
import subprocess  # noqa: S404
import typing

import pytest

from . import util
from .util import test_cfg  # noqa: F401  # pytest fixture


if typing.TYPE_CHECKING:
    import pathlib
    from typing import Final

    from testsigs import defs


def test_list(test_cfg: util.ConfigAggr) -> None:  # noqa: F811  # pytest fixture
    """Make sure `debsigs --list` on an unsigned package does not return anything."""
    with test_cfg.cfg_with_tempd() as cfg:
        if not test_cfg.has_feature("cmd-list", "1"):
            pytest.skip("No `debsigs --list` support")

        assert util.debsigs_list(cfg, test_cfg.changes_orig.debs[0]) == []


def _copy_and_sign_package(cfg: defs.Config, pkg_orig: pathlib.Path, sigtype: str) -> pathlib.Path:
    """Copy the package file, run `debsigs --sign`."""
    pkg_signed: Final = cfg.path.work / pkg_orig.name
    shutil.copy2(pkg_orig, pkg_signed)

    subprocess.check_call(  # noqa: S603
        [cfg.prog.debsigs, f"--sign={sigtype}", "--", pkg_signed.name],
        cwd=cfg.path.work,
        env=cfg.env,
    )

    cmp_res: Final = subprocess.run(  # noqa: S603
        ["cmp", "--", pkg_orig, pkg_signed],  # noqa: S607
        check=False,
        cwd=cfg.path.work,
        env=cfg.env,
    )
    if cmp_res.returncode == 0:
        raise RuntimeError(repr((cfg.prog.debsigs, pkg_orig, pkg_signed)))

    return pkg_signed


def _test_ar_sections(cfg: defs.Config, pkg_orig: pathlib.Path, pkg_signed: pathlib.Path) -> None:
    """Make sure the signed Debian package has exactly one more archive section."""
    sections_orig: Final = subprocess.check_output(  # noqa: S603
        ["ar", "t", "--", pkg_orig],  # noqa: S607
        cwd=cfg.path.work,
        encoding="UTF-8",
        env=cfg.env,
    ).splitlines()
    if any(name.startswith("_gpg") for name in sections_orig):
        raise RuntimeError(repr(sections_orig))

    sections_signed: Final = subprocess.check_output(  # noqa: S603
        ["ar", "t", "--", pkg_signed],  # noqa: S607
        cwd=cfg.path.work,
        encoding="UTF-8",
        env=cfg.env,
    ).splitlines()
    if sections_signed != [*sections_orig, "_gpgorigin"]:
        raise RuntimeError(repr((sections_orig, sections_signed)))


def _test_with_debsigs(tcfg: util.ConfigAggr, cfg: defs.Config, pkg_signed: pathlib.Path) -> None:
    """Run `debsigs --verify` now that it has actually been implemented."""
    if not tcfg.has_feature("cmd-verify", "1"):
        pytest.skip("No `debsigs --verify` support")

    subprocess.check_call(  # noqa: S603
        [cfg.prog.debsigs, "--verify", pkg_signed],
        cwd=cfg.path.work,
        env=cfg.env,
    )


def _test_with_debsig_verify(
    cfg: defs.Config,
    keys: defs.KeyFiles,
    pkg_signed: pathlib.Path,
    sign_fpr: str,
) -> None:
    """Prepare the policy file for `debsig-verify` and run it."""
    # Copied over from debsig-verify's `test/policies/*/generic.pol`
    policy_text: Final = f"""<?xml version="1.0"?>
<!DOCTYPE Policy SYSTEM "https://www.debian.org/debsig/1.0/policy.dtd">
<Policy xmlns="https://www.debian.org/debsig/1.0/">

  <!-- This is mainly a sanity check, since our filename is that of the ID
       anyway. -->
  <Origin Name="origin" id="{sign_fpr}" Description="Debsig testing"/>

  <!-- This is required to match in order for this policy to be used. We
       reject the release Type, since we want a different rule set for
       that. -->
  <Selection>
    <Required Type="origin" File="{keys.public_bin.name}" id="{sign_fpr}"/>
  </Selection>

  <!-- Once we decide to use this policy, this must pass in order to verify
       the package. -->
  <Verification MinOptional="0">
    <Required Type="origin" File="{keys.public_bin.name}" id="{sign_fpr}"/>
  </Verification>

</Policy>
"""
    print(policy_text)  # noqa: T201

    poldir: Final = cfg.path.work / "policies"
    poldir.mkdir(mode=0o755, exist_ok=True)
    polsub: Final = poldir / sign_fpr
    polsub.mkdir(mode=0o755, exist_ok=True)
    (polsub / "origin.pol").write_text(policy_text, encoding="UTF-8")

    shutil.copy2(keys.public_bin, polsub)

    subprocess.check_call(  # noqa: S603
        [
            cfg.prog.debsig_verify,
            "-d",
            "--policies-dir",
            poldir,
            "--keyrings-dir",
            poldir,
            "--list-policies",
            pkg_signed,
        ],
        cwd=cfg.path.work,
        env=cfg.env,
    )

    subprocess.check_call(  # noqa: S603
        [
            cfg.prog.debsig_verify,
            "-d",
            "--policies-dir",
            poldir,
            "--keyrings-dir",
            poldir,
            pkg_signed,
        ],
        cwd=cfg.path.work,
        env=cfg.env,
    )


def test_sign_and_verify(test_cfg: util.ConfigAggr) -> None:  # noqa: F811  # pytest fixture
    """Prepare the working directory, run the tests."""
    if not test_cfg.has_all_features({"cmd-list": "1", "cmd-sign": "1", "i-gpg": "1"}):
        pytest.skip("No `debsigs --list`, `--sign`, or GnuPG support")

    with test_cfg.cfg_with_tempd() as cfg:
        keys: Final = test_cfg.keys
        pub_key: Final = test_cfg.pub_key
        pkg_orig: Final = test_cfg.changes_orig.debs[0]

        sign_key: Final = util.get_signing_subkey(pub_key)

        sigs_empty: Final = util.debsigs_list(cfg, pkg_orig)
        if sigs_empty:
            raise RuntimeError(repr((pkg_orig, sigs_empty)))

        pkg_signed: Final = _copy_and_sign_package(cfg, pkg_orig, "origin")
        sigs_single_origin: Final = util.debsigs_list(cfg, pkg_signed)
        match sigs_single_origin:
            case [single] if single.sig_type == "origin" and single.key_id == sign_key.key_id:
                pass

            case _:
                raise RuntimeError(repr((pkg_signed, sigs_single_origin, pub_key)))

        _test_ar_sections(cfg, pkg_orig, pkg_signed)
        _test_with_debsigs(test_cfg, cfg, pkg_signed)
        _test_with_debsig_verify(cfg, keys, pkg_signed, sign_key.fpr)



# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: GPL-2.0-or-later
"""Run a stateless OpenPGP command-line tool."""

from __future__ import annotations

import subprocess  # noqa: S404
import typing

from testsigs.parse import gnupg as p_gnupg

from . import base


if typing.TYPE_CHECKING:
    import pathlib
    from typing import Final

    from testsigs import defs


def _gnupg_list_keys(cfg: defs.Config) -> list[defs.PublicKey]:
    """Run `gpg --list-keys` and partially parse the output."""
    list_output: Final = subprocess.check_output(  # noqa: S603
        [cfg.prog.gnupg, "--batch", "--with-colons", "--list-keys"],
        cwd=cfg.path.work,
        encoding="UTF-8",
        env=cfg.env,
    )
    print(list_output)  # noqa: T201
    return p_gnupg.parse_list_keys(list_output)


class GnuPGCli(base.PGPCli):
    """Invoke GnuPG as far as we can in batch mode."""

    def generate_keys(self, uid: str) -> defs.KeyFiles:
        """Generate an OpenPGP key to sign with."""
        # GnuPG makes it... difficult to generate keys in an unattended manner.
        raise NotImplementedError

    def import_secret_key(self, keys: defs.KeyFiles, uid: str) -> defs.PublicKey:
        """Import a secret key, return the parts of the public key we care about."""
        no_keys: Final = _gnupg_list_keys(self.cfg)
        if no_keys:
            raise RuntimeError(repr((self.cfg.prog.gnupg, no_keys)))

        subprocess.check_call(  # noqa: S603
            [self.cfg.prog.gnupg, "--import", keys.secret_bin],
            cwd=self.cfg.path.work,
            env=self.cfg.env,
        )

        single_key: Final = _gnupg_list_keys(self.cfg)
        if len(single_key) != 1 or single_key[0].uid != uid:
            raise RuntimeError(repr((self.cfg.prog.gnupg, single_key)))
        return single_key[0]

    def sign_detached(self, _keys: defs.KeyFiles, datafile: pathlib.Path) -> pathlib.Path:
        """Create an ASCII-armored detached signature, return the path to the new file."""
        signature: Final = datafile.with_name(datafile.name + ".asc")
        subprocess.check_call(  # noqa: S603
            [self.cfg.prog.gnupg, "-ab", datafile.name],
            cwd=self.cfg.path.work,
            env=self.cfg.env,
        )
        if not signature.is_file():
            raise RuntimeError(repr((datafile, signature)))
        return signature

    def verify_detached(
        self,
        _keys: defs.KeyFiles,
        _pub_key: defs.PublicKey,
        datafile: pathlib.Path,
        signature: pathlib.Path,
    ) -> None:
        """Verify an OpenPGP detached signature in a file."""
        subprocess.check_call(  # noqa: S603
            [self.cfg.prog.gnupg, "--verify", signature, datafile],
            cwd=self.cfg.path.work,
            env=self.cfg.env,
        )



# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: GPL-2.0-or-later
"""Unit tests for the `debsigs-installer` tool."""

from __future__ import annotations

import typing

import pytest

from .util import test_cfg  # noqa: F401  # pytest fixture


if typing.TYPE_CHECKING:
    from . import util


def test_installer(test_cfg: util.ConfigAggr) -> None:  # noqa: F811  # pytest fixture
    """Sign files, copy them over to an archive tree."""
    if not test_cfg.has_feature("tool-installer", "1"):
        pytest.skip("No `debsigs-installer` at feature version 1 support")

    raise NotImplementedError



# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: GPL-2.0-or-later
"""Unit tests for the `debsigs-signchanges` tool."""

from __future__ import annotations

import subprocess  # noqa: S404
import typing

import pytest

from . import util
from .util import test_cfg  # noqa: F401  # pytest fixture


if typing.TYPE_CHECKING:
    from typing import Final


def test_sign(test_cfg: util.ConfigAggr) -> None:  # noqa: F811  # pytest fixture
    """Sign files in different ways."""
    if not test_cfg.has_all_features(
        {"cmd-list": "1", "cmd-sign": "1", "i-gpg": "1", "tool-signchanges": "0.1"},
    ):
        pytest.skip("No `debsigs-signchanges` or `debsigs --list` support")

    with test_cfg.cfg_with_tempd() as cfg:
        pub_key: Final = test_cfg.pub_key
        changes_orig: Final = test_cfg.changes_orig

        changes_signed: Final = changes_orig.copy_to(cfg.path.work)
        assert changes_signed.get_different_files(cfg, changes_orig) == []
        for deb in changes_signed.debs:
            assert util.debsigs_list(cfg, deb) == []

        sign_key: Final = util.get_signing_subkey(pub_key)
        env_with_path: Final = dict(cfg.env)
        env_with_path["PATH"] = f"{cfg.prog.debsigs.parent}:{env_with_path['PATH']}"
        if "DEBSIGSSIG" in env_with_path:
            del env_with_path["DEBSIGSSIG"]
        env_with_path["DEBSIGSOPTIONS"] = f"--default-key={sign_key.key_id}"

        subprocess.check_call(  # noqa: S603
            [cfg.prog.debsigs.parent / "debsigs-signchanges", changes_signed.path],
            cwd=cfg.path.work,
            env=env_with_path,
        )

        changes_signed.verify(cfg)
        assert changes_signed.get_different_files(cfg, changes_orig) == sorted(
            deb.name for deb in changes_signed.debs
        )

        for deb in changes_signed.debs:
            lines_maint_only = util.debsigs_list(cfg, deb)
            match lines_maint_only:
                case [single] if single.sig_type == "maint" and single.key_id == sign_key.key_id:
                    pass

                case _:
                    raise AssertionError(repr((changes_signed, deb, lines_maint_only)))



# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: GPL-2.0-or-later
"""Test the `debsigs` command-line tool using `debsig-verify` and OpenPGP tools."""

from __future__ import annotations

import contextlib
import dataclasses
import os
import pathlib
import subprocess  # noqa: S404
import sys
import tempfile
import typing

import click
import tomli_w
import typedload

from . import defs
from .pgpcli import gnupg as c_gnupg
from .pgpcli import sop as c_sop


if typing.TYPE_CHECKING:
    from collections.abc import Iterator
    from typing import Final


OPGP_UID: Final = "Kieron Latimer <kieron@example.net>"
"""The user ID to generate a test key with."""

DEB_NAME_1: Final = "single-file"
"""The name of the first Debian package file to build and sign."""

DEB_NAME_2: Final = "more-files"
"""The name of the second Debian package file to build and sign."""


@contextlib.contextmanager
def _prepare_temp_dir(cfg: defs.Config) -> Iterator[defs.Config]:
    """Create a temporary directory, clean it up afterwards."""
    with tempfile.TemporaryDirectory(prefix="debsigs-functional.") as tempd_obj:
        tempd: Final = pathlib.Path(tempd_obj)
        home: Final = tempd / "home"
        work: Final = tempd / "work"

        home.mkdir(mode=0o755)
        work.mkdir(mode=0o755)

        env: Final = dict(cfg.env)
        env["LC_ALL"] = "C.UTF-8"
        env["LANGUAGE"] = ""
        env["HOME"] = str(home)
        for name in (name for name in env if name.startswith(("GNUPG", "GPG"))):
            del env[name]

        yield dataclasses.replace(
            cfg,
            env=env,
            path=defs.ConfigPaths(base=tempd, home=home, work=work, data=cfg.path.data),
        )


def _test_sign_file(
    cfg: defs.Config,
    keys: defs.KeyFiles,
    cli_gnupg: c_gnupg.GnuPGCli,
    cli_sop: c_sop.SOPCli,
    pub_key: defs.PublicKey,
) -> None:
    """Create a temporary file, sign it, verify the signature."""
    contents: Final = "This is a test.\n"
    testf: Final = cfg.path.work / "sign-test.txt"

    if testf.exists() or testf.is_symlink():
        raise RuntimeError(repr(testf))

    try:
        testf.write_text(contents, encoding="UTF-8")
        signature: Final = cli_gnupg.sign_detached(keys, testf)
        if signature.parent != testf.parent:
            raise RuntimeError(repr((testf, signature)))

        cli_sop.verify_detached(keys, pub_key, testf, signature)
        signature.unlink()

        det_signature: Final = cli_sop.sign_detached(keys, testf)
        if det_signature != signature:
            raise RuntimeError(repr((signature, det_signature)))
        cli_gnupg.verify_detached(keys, pub_key, testf, signature)
    finally:
        if signature.exists():
            signature.unlink()
        if testf.exists():
            testf.unlink()


def _get_deb_changes(cfg: defs.Config) -> defs.ChangesFile:
    """Fetch a Debian package from the configured Apt repositories."""
    ppath: Final = cfg.path.home / "deb-orig"
    if not ppath.is_dir():
        ppath.mkdir(mode=0o755)

    current_files: Final = {path for path in ppath.iterdir() if path.is_file()}
    for dirname in (DEB_NAME_1, DEB_NAME_2):
        subprocess.check_call(  # noqa: S603
            ["dpkg-deb", "-b", cfg.path.data / "dpkg" / dirname, ppath],  # noqa: S607
            cwd=cfg.path.work,
            env=cfg.env,
        )
    all_files: Final = {path for path in ppath.iterdir() if path.is_file()}
    new_files: Final = list(all_files - current_files)
    gone_files: Final = list(current_files - all_files)
    match (new_files, gone_files):
        case ([_first, _second], []):
            pass

        case _:
            raise RuntimeError(repr((current_files, all_files)))

    others: Final = [ppath / "debsigs-test-more.txt"]
    others[0].write_text("This file should not be changed.\n", encoding="UTF-8")

    changes: Final = defs.ChangesFile.craft(cfg, ppath / "debsigs-test.changes", new_files, others)
    changes.verify(cfg)
    return changes


def _store_test_config(
    cfg: defs.Config,
    keys: defs.KeyFiles,
    pub_key: defs.PublicKey,
    changes_orig: defs.ChangesFile,
) -> dict[str, str]:
    """Dump the configuration into TOML files, store the paths into the environment."""
    cfgpath: Final = cfg.path.work / "test-config.toml"
    with cfgpath.open(mode="wb") as cfgf:
        tomli_w.dump(
            typedload.dump(
                {
                    "format": {"version": {"major": 1, "minor": 0}},
                    "cfg": cfg,
                    "keys": keys,
                    "pub_key": pub_key,
                    "changes_orig": changes_orig,
                },
            ),
            cfgf,
        )

    env_test: Final = dict(cfg.env)
    env_test["DEBSIGS_TEST_CONFIG_FILE"] = str(cfgpath)
    return env_test


def run_test(cfg: defs.Config) -> None:
    """Prepare the working directory, run the tests."""
    cli_gnupg: Final = c_gnupg.GnuPGCli(cfg)
    cli_sop: Final = c_sop.SOPCli(cfg)

    keys: Final = cli_sop.generate_keys(OPGP_UID)
    pub_key: Final = cli_gnupg.import_secret_key(keys, OPGP_UID)
    _test_sign_file(cfg, keys, cli_gnupg, cli_sop, pub_key)

    changes_orig: Final = _get_deb_changes(cfg)

    env_test: Final = _store_test_config(cfg, keys, pub_key, changes_orig)
    subprocess.check_call(  # noqa: S603
        [
            sys.executable,
            "-m",
            "pytest",
            "--pyargs",
            "-vv",
            "testsigs.unit",
        ],
        cwd=cfg.path.work,
        env=env_test,
    )


@click.command(name="functional")
@click.option(
    "--debsigs-program",
    type=click.Path(exists=True, executable=True, path_type=pathlib.Path, resolve_path=True),
    default="/usr/bin/debsigs",
    help="the debsigs program to test",
)
@click.option(
    "--debsig-verify-program",
    type=click.Path(exists=True, executable=True, path_type=pathlib.Path, resolve_path=True),
    default="/usr/bin/debsig-verify",
    help="the debsig-verify program to check the signed package files with",
)
@click.option(
    "--gnupg-program",
    type=click.Path(exists=True, executable=True, path_type=pathlib.Path, resolve_path=True),
    default="/usr/bin/gpg",
    help="the GnuPG program to test and sign with",
)
@click.option(
    "--sop-program",
    type=click.Path(exists=True, executable=True, path_type=pathlib.Path, resolve_path=True),
    default="/usr/bin/sqop",
    help="the stateless OpenPGP program to generate keys with",
)
@click.option(
    "--test-datadir",
    type=click.Path(
        exists=True,
        file_okay=False,
        dir_okay=True,
        path_type=pathlib.Path,
        resolve_path=True,
    ),
    default="test-data",
    help="the path to the directory containing the test data files",
)
def main(
    *,
    debsigs_program: pathlib.Path,
    debsig_verify_program: pathlib.Path,
    gnupg_program: pathlib.Path,
    sop_program: pathlib.Path,
    test_datadir: pathlib.Path,
) -> None:
    """Parse command-line options, prepare the working directory, run the tests."""
    base_cfg: Final = defs.Config(
        env=dict(os.environ),
        path=defs.ConfigPaths(
            base=pathlib.Path(),
            home=pathlib.Path(),
            work=pathlib.Path(),
            data=test_datadir,
        ),
        prog=defs.ConfigPrograms(
            debsigs=debsigs_program,
            debsig_verify=debsig_verify_program,
            gnupg=gnupg_program,
            sop=sop_program,
        ),
    )
    with _prepare_temp_dir(base_cfg) as cfg:
        run_test(cfg)


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: GPL-2.0-or-later
"""Helper functions for the `testsigs` unit test suite for `debsigs."""

from __future__ import annotations

import contextlib
import dataclasses
import functools
import itertools
import os
import pathlib
import subprocess  # noqa: S404
import sys
import tempfile
import typing

import feature_check
import pytest
import typedload.dataloader
from feature_check import version as fver

from testsigs import defs  # noqa: TC001  # typedload needs this
from testsigs.parse import debsigs as p_debsigs


if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib


if typing.TYPE_CHECKING:
    from collections.abc import Iterator
    from typing import Final


@dataclasses.dataclass(frozen=True)
class ConfigAggrBase:
    """The aggregated test configuration stored by the `testsigs` runner."""

    cfg: defs.Config
    """The runtime configuration of the `testsigs` runner."""

    keys: defs.KeyFiles
    """The OpenPGP keys generated beforehand."""

    pub_key: defs.PublicKey
    """The OpenPGP key to sign with."""

    changes_orig: defs.ChangesFile
    """The path to the downloaded Debian package to sign."""

    @contextlib.contextmanager
    def cfg_with_tempd(self) -> Iterator[defs.Config]:
        """Create a temporary directory, replace `cfg.path.work` with that."""
        with tempfile.TemporaryDirectory(prefix="unit.", dir=self.cfg.path.work) as tempd_obj:
            tempd: Final = pathlib.Path(tempd_obj)
            yield dataclasses.replace(
                self.cfg,
                path=dataclasses.replace(self.cfg.path, work=tempd),
            )


@dataclasses.dataclass(frozen=True)
class ConfigAggr(ConfigAggrBase):
    """The aggregated test configuration along with some more things we may need."""

    debsigs_features: dict[str, feature_check.Version]
    """The parsed output of `debsigs --features`."""

    def has_feature(self, feature: str, min_version: str) -> bool:
        """Determine whether the specified feature is there at that version or later."""
        found: Final = self.debsigs_features.get(feature)
        if found is None:
            return False

        return fver.version_compare(found, feature_check.parse_version(min_version)) >= 0

    def has_all_features(self, features: dict[str, str]) -> bool:
        """Determine whether all of the specified features are there at least at those versions."""
        return all(itertools.starmap(self.has_feature, features.items()))


@functools.lru_cache
def get_loader() -> typedload.dataloader.Loader:
    """Prepare a dataloader that understands deferred type annotations."""
    return typedload.dataloader.Loader(pep563=True)


@pytest.fixture
def test_cfg() -> ConfigAggr:
    """Load the aggregated test configuration."""
    with pathlib.Path(os.environ["DEBSIGS_TEST_CONFIG_FILE"]).open(mode="rb") as cfgf:
        cfg_base: Final = get_loader().load(tomllib.load(cfgf), ConfigAggrBase)

    return ConfigAggr(
        cfg=cfg_base.cfg,
        keys=cfg_base.keys,
        pub_key=cfg_base.pub_key,
        changes_orig=cfg_base.changes_orig,
        debsigs_features=feature_check.obtain_features(str(cfg_base.cfg.prog.debsigs)),
    )


def debsigs_list(cfg: defs.Config, pkg: pathlib.Path) -> list[defs.DebSig]:
    """Run `debsigs --list`, return any signatures returned."""
    raw: Final = subprocess.check_output(  # noqa: S603
        [cfg.prog.debsigs, "--list", "--", pkg],
        cwd=cfg.path.work,
        encoding="UTF-8",
        env=cfg.env,
    )
    return p_debsigs.parse_list(raw)


def get_signing_subkey(pub_key: defs.PublicKey) -> defs.Subkey:
    """Find a single subkey capable of signing."""
    match [subkey for subkey in pub_key.subkeys if "s" in subkey.capabilities]:
        case [single]:
            return single

        case _:
            raise RuntimeError(repr(pub_key))



# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: GPL-2.0-or-later
"""Parse the output of `gpg --list-keys --with-colons`."""

from __future__ import annotations

import dataclasses
import typing

import pyparsing as pyp

from testsigs import defs


if typing.TYPE_CHECKING:
    from typing import Final


_p_eol = pyp.Char("\r\n").suppress()

_p_ign = pyp.ZeroOrMore(pyp.CharsNotIn(":\n")).suppress() + pyp.Char(":").suppress()

_p_ign_all = pyp.CharsNotIn("\n").suppress()

_p_hex_string = pyp.Word("0123456789ABCDEF")

_p_string = pyp.CharsNotIn(":\n")

_p_ignored_line = pyp.Literal("tru:").suppress() + _p_ign_all + _p_eol

_p_ignored_lines = pyp.ZeroOrMore(_p_ignored_line)

_p_list_pub = (
    pyp.Literal("pub:").suppress() + _p_ign + _p_ign + _p_ign + _p_hex_string + _p_ign_all + _p_eol
)

_p_list_fpr = pyp.Literal("fpr:").suppress() + _p_ign * 8 + _p_hex_string + _p_ign_all + _p_eol

_p_list_uid = pyp.Literal("uid:").suppress() + _p_ign * 8 + _p_string + _p_ign_all + _p_eol

_p_list_sub = (
    pyp.Literal("sub:").suppress()
    + _p_ign
    + _p_ign
    + _p_ign
    + _p_hex_string
    + _p_ign * 7
    + _p_string
    + _p_ign_all
    + _p_eol
)


@_p_list_sub.set_parse_action
def _parse_list_sub(tokens: pyp.ParseResults) -> defs.Subkey:
    """Parse a subkey definition."""
    params: Final = tokens.as_list()
    match params:
        case [key_id, caps]:
            return defs.Subkey(key_id=key_id, fpr="", capabilities=set(caps))

        case _:
            raise RuntimeError(repr(params))


_p_list_subkey = _p_ignored_lines + _p_list_sub + _p_ignored_lines + _p_list_fpr


@_p_list_subkey.set_parse_action
def _parse_list_subkey(tokens: pyp.ParseResults) -> defs.Subkey:
    """Parse a subkey definition."""
    params: Final = tokens.as_list()
    match params:
        case [sub_key_id, fpr]:
            return dataclasses.replace(sub_key_id, fpr=fpr)

        case _:
            raise RuntimeError(repr(params))


_p_list_pubkey = (
    _p_ignored_lines
    + _p_list_pub
    + _p_ignored_lines
    + _p_list_fpr
    + _p_ignored_lines
    + _p_list_uid
    + pyp.OneOrMore(_p_list_subkey)
)


@_p_list_pubkey.set_parse_action
def _parse_list_pubkey(tokens: pyp.ParseResults) -> defs.PublicKey:
    """Parse a subkey definition."""
    params: Final = tokens.as_list()
    match params:
        case [key_id, fpr, uid, *subkeys] if isinstance(key_id, str) and isinstance(
            fpr,
            str,
        ) and isinstance(uid, str) and subkeys and all(
            isinstance(subkey, defs.Subkey) for subkey in subkeys
        ):
            # We did just validate that one...
            return defs.PublicKey(
                key_id=key_id,
                fpr=fpr,
                uid=uid,
                subkeys=subkeys,  # type: ignore[arg-type]
            )

        case _:
            raise RuntimeError(repr(params))


_p_list_keys = pyp.ZeroOrMore(_p_list_pubkey) + _p_ignored_lines

_p_list_keys_complete = _p_list_keys.leave_whitespace().parse_with_tabs()


def parse_list_keys(output: str) -> list[defs.PublicKey]:
    """Parse the output of `gpg --list-keys`, extract the info we care about."""
    res: Final = _p_list_keys_complete.parse_string(output, parse_all=True).as_list()
    if not all(isinstance(pkey, defs.PublicKey) for pkey in res):
        raise RuntimeError(repr(res))
    return res



# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: GPL-2.0-or-later
"""Parse the human-readable output of `debsigs --list`."""

from __future__ import annotations

import typing

import pyparsing as pyp

from testsigs import defs


if typing.TYPE_CHECKING:
    from typing import Final


_p_first = pyp.Literal("GPG signatures in ") + pyp.CharsNotIn("\n") + pyp.Char("\n")

_p_sig_type = pyp.Regex("[A-Za-z0-9_]+")

_p_hex = pyp.Word("0123456789ABCDEF")

_p_sig = (
    _p_sig_type
    + pyp.Literal(": signed by ").suppress()
    + _p_hex
    + pyp.Literal(" on ").suppress()
    + pyp.CharsNotIn("\n").suppress()
    + pyp.Char("\n").suppress()
)


@_p_sig.set_parse_action
def _parse_sig(tokens: pyp.ParseResults) -> defs.DebSig:
    """Parse a signature line."""
    match tokens.as_list():
        case [sig_type, key_id] if isinstance(sig_type, str) and isinstance(key_id, str):
            return defs.DebSig(sig_type=sig_type, key_id=key_id)

        case _:
            raise RuntimeError(repr(tokens))


_p_last = pyp.Literal(" *** NO ATTEMPT HAS BEEN MADE") + pyp.CharsNotIn("\n") + pyp.Char("\n")

_p_list = _p_first.suppress() + pyp.ZeroOrMore(_p_sig) + _p_last.suppress()

_p_list_complete = _p_list.leave_whitespace().parse_with_tabs()


def parse_list(output: str) -> list[defs.DebSig]:
    """Parse the output of `debsigs --list`, extract the info we care about."""
    res: Final = _p_list_complete.parse_string(output, parse_all=True).as_list()
    if not all(isinstance(sig, defs.DebSig) for sig in res):
        raise RuntimeError(repr(res))
    return res



# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: GPL-2.0-or-later
"""Common definitions for the `debsigs` functional test suite."""

from __future__ import annotations

import dataclasses
import pathlib  # noqa: TC003  # typedload needs this
import shutil
import subprocess  # noqa: S404
import typing


if typing.TYPE_CHECKING:
    from typing import Final, TypeVar

    TChangesFile = TypeVar("TChangesFile", bound="ChangesFile")


@dataclasses.dataclass
class Error(Exception):
    """The base class for errors that occurred during the debsigs test."""


@dataclasses.dataclass(frozen=True)
class ConfigPaths:
    """The paths to the temporary directories to run the test in."""

    base: pathlib.Path
    """The base directory, the temporary directory created for the whole test."""

    home: pathlib.Path
    """The directory to serve as the current account's home directory."""

    work: pathlib.Path
    """The directory to run the test in."""

    data: pathlib.Path
    """The directory containing the data files to test with, e.g. Debian package structure."""


@dataclasses.dataclass(frozen=True)
class ConfigPrograms:
    """The paths to the programs to test and test with."""

    debsigs: pathlib.Path
    """The path to the `debsigs` implementation to test."""

    debsig_verify: pathlib.Path
    """The path to the `debsig-verify` program to test the signed files with."""

    gnupg: pathlib.Path
    """The path to the `gpg` program to use and test with."""

    sop: pathlib.Path
    """The path to the stateless OpenPGP program to create keys and possibly sign with."""


@dataclasses.dataclass(frozen=True)
class Config:
    """Runtime configuration for the test suite."""

    env: dict[str, str]
    """The environment variables to run child processes with."""

    path: ConfigPaths
    """The paths to the temporary directories to run the test in."""

    prog: ConfigPrograms
    """The paths to the programs to test and test with."""


@dataclasses.dataclass(frozen=True)
class KeyFiles:
    """The paths to the armored and dearmored OpenPGP key files to test with."""

    secret: pathlib.Path
    """The ASCII-armored secret key file."""

    secret_bin: pathlib.Path
    """The dearmored secret key file."""

    public: pathlib.Path
    """The ASCII-armored public key file."""

    public_bin: pathlib.Path
    """The dearmored public key file."""


@dataclasses.dataclass(frozen=True)
class Subkey:
    """The parts of an OpenPGP subkey that we care about."""

    key_id: str
    """The hex string identifier of the subkey."""

    fpr: str
    """The full hex fingerprint of the key."""

    capabilities: set[str]
    """The capabilities of this subkey (e.g. `s` for sign, `e` for encrypt, etc.)."""


@dataclasses.dataclass(frozen=True)
class PublicKey:
    """The parts of an OpenPGP public key that we care about."""

    key_id: str
    """The hex string identifier of the public key."""

    fpr: str
    """The full hex fingerprint of the key."""

    uid: str
    """The user ID associated with this key."""

    subkeys: list[Subkey]
    """The subkeys defined for this key (there must be at least one)."""


@dataclasses.dataclass(frozen=True)
class ChangesFile:
    """List the files described in a Debian archive `*.changes` file."""

    path: pathlib.Path
    """The path to the changes file itself."""

    debs: list[pathlib.Path]
    """The Debian packages included in the changes file."""

    others: list[pathlib.Path]
    """Any other files included in the changes file."""

    @classmethod
    def craft(
        cls: type[TChangesFile],
        cfg: Config,
        path: pathlib.Path,
        debs: list[pathlib.Path],
        others: list[pathlib.Path],
    ) -> TChangesFile:
        """Create something resembling a `*.changes` file."""
        if any(item.parent != path.parent or item == path for item in debs):
            raise RuntimeError(repr((path, debs)))
        if any(item.parent != path.parent or item == path for item in others):
            raise RuntimeError(repr((path, others)))
        if set(debs) & set(others):
            raise RuntimeError(repr((debs, others)))

        binaries: Final = [deb.name.split("_")[0] for deb in debs]
        all_files: Final = [*debs, *others]
        contents = f"""Format: 1.8
Date: Wed, 29 Dec 2021 16:46:40 +0200
Source: debsigs-test
Binary: {" ".join(binaries)}
Architecture: source all
Version: 0.1.0
Distribution: unstable
Urgency: medium
Maintainer: Peter Pentchev <roam@debian.org>
Changed-By: Peter Pentchev <roam@debian.org>
Description:
 debsigs-test    - test the operation of the debsigs suite
Changes:
 debsigs-test (0.1.0) unstable; urgency=medium
 .
   * Nothing really.
"""

        for heading in ("Sha1", "Sha256"):
            contents += f"Checksums-{heading}:\n"
            for filepath in all_files:
                checksum = subprocess.check_output(  # noqa: S603
                    [
                        f"{heading.lower()}sum",
                        "--",
                        filepath,
                    ],
                    encoding="UTF-8",
                    env=cfg.env,
                ).split()[0]
                filesize = filepath.stat().st_size
                contents += f" {checksum} {filesize} {filepath.name}\n"

        contents += "Files:\n"
        for filepath in all_files:
            checksum = subprocess.check_output(  # noqa: S603
                [  # noqa: S607
                    "md5sum",
                    "--",
                    filepath,
                ],
                cwd=cfg.path.work,
                encoding="UTF-8",
                env=cfg.env,
            ).split()[0]
            filesize = filepath.stat().st_size
            contents += f" {checksum} {filesize} devel optional {filepath.name}\n"

        path.write_text(contents, encoding="UTF-8")
        return cls(path=path, debs=debs, others=others)

    def verify(self, cfg: Config) -> None:
        """Run `dscverify` without verifying the nonexistent OpenPGP signature."""
        lines: Final = subprocess.check_output(  # noqa: S603
            ["dscverify", "-u", "--", self.path],  # noqa: S607
            cwd=cfg.path.work,
            encoding="UTF-8",
            env=cfg.env,
        ).splitlines()
        verified: Final = {
            part[2] for part in (line.partition("validating ") for line in lines) if part[1]
        }
        if verified != {deb.name for deb in self.debs} | {other.name for other in self.others}:
            raise RuntimeError(repr((self, lines)))

    def copy_to(self: TChangesFile, destdir: pathlib.Path) -> TChangesFile:
        """Copy the changes file and all the files it references over to another location."""
        npath: Final = destdir / self.path.name
        shutil.copy2(self.path, npath)

        ndebs: Final = []
        for deb in self.debs:
            ndeb = destdir / deb.name
            shutil.copy2(deb, ndeb)
            ndebs.append(ndeb)

        nothers: Final = []
        for other in self.others:
            nother = destdir / other.name
            shutil.copy2(other, nother)
            nothers.append(nother)

        return type(self)(path=npath, debs=ndebs, others=nothers)

    def get_different_files(self, cfg: Config, other: ChangesFile) -> list[str]:
        """Get the filenames of the files that differ between this changes file and another one."""
        ours: Final = {path.name: path for path in [*self.debs, *self.others]}
        theirs: Final = {path.name: path for path in [*other.debs, *other.others]}
        our_names: Final = set(ours)
        their_names: Final = set(theirs)
        return sorted(
            our_names.symmetric_difference(their_names)
            | {
                name
                for name in (our_names & their_names)
                if subprocess.run(  # noqa: S603
                    ["cmp", "--", ours[name], theirs[name]],  # noqa: S607
                    check=False,
                    cwd=cfg.path.work,
                    env=cfg.env,
                ).returncode
                != 0
            },
        )


@dataclasses.dataclass(frozen=True)
class DebSig:
    """A single signature on a package."""

    sig_type: str
    """The type of the signature, e.g. "origin", "maint", etc."""

    key_id: str
    """The hex identifier of the OpenPGP key that the signature was made with."""



# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: GPL-2.0-or-later
"""Build `debsigs` in a temporary directory, run the test."""

from __future__ import annotations

import enum
import os
import pathlib
import shlex
import subprocess  # noqa: S404
import sys
import tempfile
import typing

import click


if typing.TYPE_CHECKING:
    from typing import Final


class Impl(str, enum.Enum):
    """The supported `debsigs` implementations."""

    PERL = "perl"

    def __str__(self) -> str:
        """Use the value itself as a string representation."""
        return self.value


def copy_source(tempd: pathlib.Path, u8env: dict[str, str]) -> pathlib.Path:
    """Copy the `debsigs` source files over to the temporary directory."""
    srcdir: Final = tempd / "debsigs"
    srcdir.mkdir(mode=0o755)

    files: Final = sorted(
        set(
            subprocess.check_output(  # noqa: S603
                [  # noqa: S607
                    "git",
                    "ls-files",
                ],
                encoding="UTF-8",
                env=u8env,
            ).splitlines(),
        ),
    )

    srcball: Final = tempd / "debsigs.tar"
    subprocess.check_call(  # noqa: S603
        [  # noqa: S607
            "tar",
            "-cf",
            srcball,
            "--",
            *files,
        ],
        env=u8env,
    )
    subprocess.check_call(  # noqa: S603
        [  # noqa: S607
            "tar",
            "-xf",
            srcball,
        ],
        cwd=srcdir,
        env=u8env,
    )
    srcball.unlink()
    return srcdir


def build_debsigs_perl(
    bindir: pathlib.Path,
    srcdir: pathlib.Path,
    u8env: dict[str, str],
) -> pathlib.Path:
    """Build the Perl implementation of `debsigs`, create a wrapper script."""
    perldir: Final = srcdir / "perl"
    subprocess.check_call(  # noqa: S603
        [  # noqa: S607
            "perl",
            "Makefile.PL",
        ],
        cwd=perldir,
        env=u8env,
    )
    subprocess.check_call(  # noqa: S603
        [  # noqa: S607
            "make",
        ],
        cwd=perldir,
        env=u8env,
    )
    blibdir: Final = perldir / "blib/lib"
    if not blibdir.is_dir():
        raise RuntimeError(repr((perldir, blibdir)))

    def create_wrapper(name: str) -> pathlib.Path:
        """Create a single wrapper for one of the Perl command-line tools."""
        wrapper: Final = bindir / name
        wrapper.write_text(
            f"""#!/bin/sh

set -e

exec perl -I{shlex.quote(str(blibdir))} -- {shlex.quote(str(perldir / name))} "$@"
""",
            encoding="UTF-8",
        )
        wrapper.chmod(0o755)
        return wrapper

    wrapper: Final = create_wrapper("debsigs")
    # Make sure the wrapper script works.
    version: Final = subprocess.check_output(  # noqa: S603
        [wrapper, "--version"],
        encoding="UTF-8",
        env=u8env,
    )
    if not version.startswith("debsigs v0.2."):
        raise RuntimeError(repr(version))

    for other in ("debsigs-autosign", "debsigs-signchanges"):
        create_wrapper(other)

    return wrapper


def build_debsigs(
    impl: Impl,
    bindir: pathlib.Path,
    srcdir: pathlib.Path,
    u8env: dict[str, str],
) -> pathlib.Path:
    """Build the `debsigs` implementation, return the path to a wrapper script."""
    match impl:
        case Impl.PERL:
            return build_debsigs_perl(bindir, srcdir, u8env)

        case _:
            raise NotImplementedError(repr(impl))


@click.command(name="build-and-test")
@click.option(
    "--debsigs-impl",
    type=Impl,
    default=Impl.PERL,
    help="the debsigs implementation to test",
)
@click.option(
    "--debsig-verify-program",
    type=click.Path(exists=True, executable=True, path_type=pathlib.Path, resolve_path=True),
    default="/usr/bin/debsig-verify",
    help="the debsig-verify program to check the signed package files with",
)
@click.option(
    "--gnupg-program",
    type=click.Path(exists=True, executable=True, path_type=pathlib.Path, resolve_path=True),
    default="/usr/bin/gpg",
    help="the GnuPG program to test and sign with",
)
@click.option(
    "--sop-program",
    type=click.Path(exists=True, executable=True, path_type=pathlib.Path, resolve_path=True),
    default="/usr/bin/sqop",
    help="the stateless OpenPGP program to generate keys with",
)
def main(
    *,
    debsigs_impl: Impl,
    debsig_verify_program: pathlib.Path,
    gnupg_program: pathlib.Path,
    sop_program: pathlib.Path,
) -> None:
    """Parse command-line options, build `debsigs`, run the test."""
    with tempfile.TemporaryDirectory(prefix="debsigs-build.") as tempd_obj:
        tempd: Final = pathlib.Path(tempd_obj)
        print(f"build-and-test: using {tempd} as a temporary directory")

        u8env: Final = dict(os.environ)
        u8env["LC_ALL"] = "C.UTF-8"
        if "LANGUAGE" in u8env:
            del u8env["LANGUAGE"]

        srcdir: Final = copy_source(tempd, u8env)
        print(f"build-and-test: copied the debsigs source to {srcdir}")

        bindir: Final = tempd / "bin"
        bindir.mkdir(mode=0o755)
        wrapper: Final = build_debsigs(debsigs_impl, bindir, srcdir, u8env)
        print(f"build-and-test: built the debsigs source, created {wrapper}")

        print("build-and-test: invoking testsigs")
        subprocess.check_call(  # noqa: S603
            [
                sys.executable,
                "-B",
                "-u",
                "-m",
                "testsigs",
                "--debsigs-program",
                wrapper,
                "--debsig-verify-program",
                debsig_verify_program,
                "--gnupg-program",
                gnupg_program,
                "--sop-program",
                sop_program,
                "--test-datadir",
                srcdir / "test-data",
            ],
            cwd=tempd,
            env=u8env,
        )

        print("build-and-test: seems fine")


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: GPL-2.0-or-later
"""Run a stateless OpenPGP command-line tool."""

from __future__ import annotations

import subprocess  # noqa: S404
import typing

from testsigs import defs

from . import base


if typing.TYPE_CHECKING:
    import pathlib
    from typing import Final


class SOPCli(base.PGPCli):
    """Invoke a stateless OpenPGP command-line tool."""

    def generate_keys(self, uid: str) -> defs.KeyFiles:
        """Generate an OpenPGP key to sign with."""
        keydir: Final = self.cfg.path.home / "keys"
        keydir.mkdir(mode=0o755)
        keys: Final = defs.KeyFiles(
            secret=keydir / "secret.asc",
            secret_bin=keydir / "secret.gpg",
            public=keydir / "public.asc",
            public_bin=keydir / "public.gpg",
        )

        secret_text: Final = subprocess.check_output(  # noqa: S603
            [self.cfg.prog.sop, "generate-key", "--profile=rfc4880", uid],
            cwd=self.cfg.path.work,
            encoding="UTF-8",
            env=self.cfg.env,
        )
        if not secret_text.startswith("-----BEGIN PGP PRIVATE KEY BLOCK-----"):
            raise RuntimeError(repr((self.cfg.prog.sop, secret_text)))
        keys.secret.write_text(secret_text, encoding="UTF-8")

        # This is not UTF-8 text; it may contain arbitrary byte values.
        secret_bin: Final = subprocess.check_output(  # noqa: S603
            [self.cfg.prog.sop, "dearmor"],
            cwd=self.cfg.path.work,
            env=self.cfg.env,
            input=secret_text.encode("UTF-8"),
        )
        keys.secret_bin.write_bytes(secret_bin)

        public_text: Final = subprocess.check_output(  # noqa: S603
            [self.cfg.prog.sop, "extract-cert"],
            cwd=self.cfg.path.work,
            encoding="UTF-8",
            env=self.cfg.env,
            input=secret_text,
        )
        if not public_text.startswith("-----BEGIN PGP PUBLIC KEY BLOCK-----"):
            raise RuntimeError(repr((self.cfg.prog.sop, public_text)))
        keys.public.write_text(public_text, encoding="UTF-8")

        # This is not UTF-8 text; it may contain arbitrary byte values.
        public_bin: Final = subprocess.check_output(  # noqa: S603
            [self.cfg.prog.sop, "dearmor"],
            cwd=self.cfg.path.work,
            env=self.cfg.env,
            input=public_text.encode("UTF-8"),
        )
        keys.public_bin.write_bytes(public_bin)

        return keys

    def import_secret_key(self, keys: defs.KeyFiles, uid: str) -> defs.PublicKey:
        """Import a secret key, return the parts of the public key we care about."""
        # We don't need that yet, and we cannot parse an OpenPGP key's structure by ourselves.
        raise NotImplementedError

    def sign_detached(self, keys: defs.KeyFiles, datafile: pathlib.Path) -> pathlib.Path:
        """Create an ASCII-armored detached signature, return the path to the new file."""
        signature: Final = datafile.with_name(f"{datafile.name}.asc")
        sig_contents: Final = subprocess.check_output(  # noqa: S603
            [self.cfg.prog.sop, "sign", keys.secret],
            cwd=self.cfg.path.work,
            env=self.cfg.env,
            input=datafile.read_bytes(),
        )
        signature.write_bytes(sig_contents)
        return signature

    def verify_detached(
        self,
        keys: defs.KeyFiles,
        pub_key: defs.PublicKey,
        datafile: pathlib.Path,
        signature: pathlib.Path,
    ) -> None:
        """Verify an OpenPGP detached signature in a file."""
        sig_report: Final = subprocess.check_output(  # noqa: S603
            [self.cfg.prog.sop, "verify", signature, keys.public],
            cwd=self.cfg.path.work,
            env=self.cfg.env,
            input=datafile.read_bytes(),
        ).decode()
        if not any(
            line.split()[2] in {pub_key.key_id, pub_key.fpr} and line.split()[3] == "mode:binary"
            for line in sig_report.splitlines()
        ):
            raise RuntimeError(repr((pub_key, sig_report)))



# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: GPL-2.0-or-later
"""Unit tests for the output of `debsigs --features`."""

from __future__ import annotations

import subprocess  # noqa: S404
import typing

from .util import test_cfg  # noqa: F401  # pytest fixture


if typing.TYPE_CHECKING:
    from typing import Final

    from . import util


def test_version(test_cfg: util.ConfigAggr) -> None:  # noqa: F811  # pytest fixture
    """Make sure `debsigs --version` and `debsigs --features` agree."""
    with test_cfg.cfg_with_tempd() as cfg:
        ver: Final = test_cfg.debsigs_features["debsigs"].value
        assert subprocess.check_output(  # noqa: S603
            [cfg.prog.debsigs, "--version"],
            cwd=cfg.path.work,
            encoding="UTF-8",
            env=cfg.env,
        ).splitlines() == [f"debsigs v{ver}"]



# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: GPL-2.0-or-later
"""The base class definition for invoking OpenPGP implementations."""

from __future__ import annotations

import abc
import typing


if typing.TYPE_CHECKING:
    import pathlib

    from testsigs import defs


class PGPCli(abc.ABC):
    """An OpenPGP implementation."""

    # Abstract dataclasses still have some quirks, so do this the hard way.

    cfg: defs.Config
    """The runtime configuration settings, e.g. the paths to the tools to invoke."""

    def __init__(self, cfg: defs.Config) -> None:
        """Initialize a `PGPCli` object."""
        self.cfg = cfg

    @abc.abstractmethod
    def generate_keys(self, uid: str) -> defs.KeyFiles:
        """Generate an OpenPGP key to sign with."""
        raise NotImplementedError

    @abc.abstractmethod
    def import_secret_key(self, keys: defs.KeyFiles, uid: str) -> defs.PublicKey:
        """Import a secret key, return the parts of the public key we care about."""
        raise NotImplementedError

    @abc.abstractmethod
    def sign_detached(self, keys: defs.KeyFiles, datafile: pathlib.Path) -> pathlib.Path:
        """Create an ASCII-armored detached signature, return the path to the new file."""
        raise NotImplementedError

    @abc.abstractmethod
    def verify_detached(
        self,
        keys: defs.KeyFiles,
        pub_key: defs.PublicKey,
        datafile: pathlib.Path,
        signature: pathlib.Path,
    ) -> None:
        """Verify an OpenPGP detached signature in a file."""
        raise NotImplementedError
